@@ -1,7 +1,7 @@
 """
 monobit - tools for working with monochrome bitmap fonts
 
-(c) 2019--2022 Rob Hagemans
+(c) 2019--2023 Rob Hagemans
 licence: https://opensource.org/licenses/MIT
 """
 
@@ -10,17 +10,18 @@ import sys as _sys
 assert _sys.version_info >= (3, 7)
 
 from .constants import VERSION as __version__
-from .pack import Pack
+from .pack import Pack, operations as _pack_operations
 from .font import Font, operations as _operations
 from .glyph import Glyph
 from . import formats
 from .storage import open_location, load, save, loaders, savers
 from .encoding import charmaps, encoder
 from .taggers import tagmaps
-from .renderer import render, to_text, to_image
+from .renderer import render
 from .labels import Char, Codepoint, Tag
 
 
+_operations.update(_pack_operations)
 # inject font operations into main module namespace
 globals().update(_operations)
 
