@@ -1,7 +1,7 @@
 """
 monobit.formats.psf - PC Screen Font format
 
-(c) 2019--2022 Rob Hagemans
+(c) 2019--2023 Rob Hagemans
 licence: https://opensource.org/licenses/MIT
 """
 
@@ -108,7 +108,7 @@ def load_psf(instream, where=None):
     else:
         raise FileFormatError(
             'Not a PSF file: '
-            f'magic bytes 0x{magic:X} not one of 0x{_PSF1_MAGIC:X}, 0x{_PSF2_MAGIC:X}'
+            f'magic bytes {magic} not one of {_PSF1_MAGIC}, {_PSF2_MAGIC}'
         )
     logging.info('PSF properties:')
     for name, value in vars(psf_props).items():
